'''
Created on 28 sept. 2014

@author: coissac
'''

from subprocess import Popen
from tempfile   import mkdtemp
from tempfile   import mktemp
from shutil     import rmtree

import atexit
import os.path
import sys

from orgasm.indexer import Index
from orgasm.reads._sequences import readPairedEnd, setLogger, getStats  # @UnresolvedImport

__title__="Index a set of reads"

default_config = { 'reformat' : False,
                   'single'   : False,
                   'forward'  : None,
                   'reverse'  : None,
                   '5trim'    : 0,
                   'qualtrim' : 0,
                   'badqual'  : 10,
                   'skip'     : 0,
                   'maxread'  : 0,
                   'length'   : 0,
                   'estimate' : 0,
                   'minlength': 0,
                   'nopipe'   : False,
                   'checkpairs':False,
                   'mate'     : False,
                   "phix"     : False,
                   "fastqdump": False,
                   'shift'    : 33,
                   "direct"   : False,
                   "zip"      : False
                 }

def addOptions(parser):
    parser.add_argument(dest='orgasm:indexfilename',  metavar='index', 
                        help='name of the produced index')

    parser.add_argument(dest='index:forward',  metavar='forward', 
                        nargs='?', 
                        default=None,
                        help='Filename of the forward reads')
    
    parser.add_argument(dest='index:reverse',     
                        metavar='reverse', 
                        nargs='?', 
                        default=None,
                        help='Filename of the reverse reads' )
    
    
    parser.add_argument("--reformat",
                        dest="index:reformat",
                        action='store_true',
                        default=None,
                        help='Asks for reformatting an old sequence index to the new format'
                       )
    
    parser.add_argument('--single',           
                        dest='index:single', 
                        action='store_true', 
                        default=None, 
                        help='Single read mode, pair-end reads will be simulated')

    parser.add_argument('--check-phiX174',           
                        dest='index:phix', 
                        action='store_true', 
                        default=None, 
                        help='Checks for PhiX174 contamination (default)')

    parser.add_argument('--no-check-phiX174',           
                        dest='index:phix', 
                        action='store_false', 
                        default=None, 
                        help='Does not check for PhiX174 contamination')

    parser.add_argument('--mate-pairs',       dest='index:mate', 
                                              action='store_true', 
                                              default=None, 
                        help='Mate pair library mode')

    parser.add_argument('--5-prime-trim',     dest='index:5trim', 
                                              type=int, 
                                              action='store', 
                                              default=None, 
                        help='Cut the N first base pairs of '
                             'reads (default %dbp)' % default_config['5trim'])
    
    parser.add_argument('--3-prime-quality',  dest='index:qualtrim', 
                                              metavar="Q",
                                              type=int, 
                                              action='store', 
                                              default=None, 
                        help="Hard clips the 3' end of each reads"
                             'after the first base with a score '
                             "less or equal to Q (default 0 no clipping)")
    
    parser.add_argument('--bad-quality',  dest='index:badqual', 
                                              metavar="Q",
                                              type=int, 
                                              action='store', 
                                              default=None, 
                        help="Consider quality below Q as bad "
                             "quality score, and try to clip "
                             "reads to maximise the overall "
                             "quality. Zero means no clipping "
                             "(default %d)" % default_config['badqual'])
    
    parser.add_argument('--skip',             dest='index:skip', 
                                              metavar="N",
                                              type=int, 
                                              action='store', 
                                              default=None, 
                        help='Skip the N first read pairs '
                             '(default %d)' % default_config['skip'])
    
    parser.add_argument('--max-reads',        dest='index:maxread', 
                                              metavar="N",
                                              type=int, 
                                              action='store', 
                                              default=None, 
                        help='index a maximum of N read pairs '
                             '(default the full file)')
    
    parser.add_argument('--length',           dest='index:length', 
                                              type=int, 
                                              action='store', 
                                              default=None, 
                        help='The length of the read to index '
                             ' (default indexed length is estimated from the first read)')
    
    parser.add_argument('--estimate-length',  dest='index:estimate', 
                                              metavar='FRACTION',
                                              type=float, 
                                              action='store', 
                                              default=None, 
                        help='Estimate the length to index for conserving FRACTION '
                             'of the overall data set')
    
    parser.add_argument('--minimum-length',   dest='index:minlength', 
                                              type=int, 
                                              action='store', 
                                              default=None, 
                        help='The minimum length of the read to index '
                             'if the --estimate-length is activated (default 81)')
    
    parser.add_argument('--no-pipe',          dest='index:nopipe', 
                                              action='store_true', 
                                              default=None, 
                        help='do not use pipe but temp files instead')

    parser.add_argument('--bypass-filtering', dest='index:direct', 
                                              action='store_true', 
                                              default=None, 
                        help='Sequence files are considered as '
                             'pre-filtered fastq files')

    parser.add_argument('--fastq-dump',           
                        dest='index:fastqdump', 
                        action='store_true', 
                        default=None, 
                        help='Dump the fastq file or the trimmed reads')

    parser.add_argument('--check-pairing',            dest='index:checkpairs', 
                                              action='store_true', 
                                              default=False, 
                        help='ensure that forward and reverse files are correctly paired')

    parser.add_argument('--low-memory',       dest='index:zip', 
                                              action='store_true', 
                                              default=None, 
                        help='Reduce memory usage for optimal length computation')

    parser.add_argument('--quality-encoding-offset',  dest='index:shift', 
                                              type=int, 
                                              action='store', 
                                              default=None, 
                        help='The code offset added to each quality '
                             'score to encode fastq quality '
                             '(default %d - Sanger format)' % default_config['shift'])
    

tmpdir = []

# @atexit.register
# def cleanup():
#     try:
#         os.unlink(FIFO)
#     except:
#         pass

def reformatOldIndex(config):

    logger  = config['orgasm']['logger']
    output  = config['orgasm']['indexfilename']
    if not (os.path.exists('%s.ofx' % output ) and
            os.path.exists('%s.ogx' % output ) and
            os.path.exists('%s.opx' % output ) and
            os.path.exists('%s.orx' % output )
           ):
        logger.error('The %s index does not exist or is not complete' % output)
        sys.exit(1)
    
    dirname = '%s.odx' % output
    if not os.path.exists(dirname):
        os.makedirs(dirname) 

    os.rename('%s.ofx' % output, '%s/index.ofx' % dirname)
    os.rename('%s.ogx' % output, '%s/index.ogx' % dirname)
    os.rename('%s.opx' % output, '%s/index.opx' % dirname)
    os.rename('%s.orx' % output, '%s/index.orx' % dirname)

    sys.exit(0)

def postponerm(filename):
    def trigger():
        print("Deleting tmp file : %s" % filename,file=sys.stderr)
        try:
            os.unlink(filename)
        except:
            pass
        
    return trigger

def postponermdir(filename):
    def trigger():
        print("Deleting tmp directory : %s" % filename,file=sys.stderr)
        try:
            rmtree(filename)
        except:
            pass
        
    return trigger

def getTmpDir():
    global tmpdir

    if not tmpdir:
        tmpdir.append(mkdtemp())
        atexit.register(postponermdir(tmpdir[0]))
        
    return tmpdir[0]


def ungzip(filename,nopipe):
    tmpdir = getTmpDir()

    fifo = mktemp(prefix="unziped-", dir=tmpdir)
    command="gzip -d -c %s > %s" % (filename,fifo)
    
    if nopipe:
        os.system(command)
    else:
        os.mkfifo(fifo)
        process = Popen(command,                                            # @UnusedVariable
                        shell=True,
                        stderr=open('/dev/null','w')) 
    
    atexit.register(postponerm(fifo))

    return fifo

def unbzip(filename,nopipe):
    tmpdir = getTmpDir()

    fifo = mktemp(prefix="unziped-", dir=tmpdir)
    command="bzip2 -d -c %s > %s" % (filename,fifo)
    atexit.register(postponerm(fifo))
    
    if nopipe:
        os.system(command)
    else:
        os.mkfifo(fifo)
        process = Popen(command,                                            # @UnusedVariable
                        shell=True,
                        stderr=open('/dev/null','w')) 
    
    return fifo
                
def formatFastq(seq):
    fastq = '@{id:0>7}\n{seq}\n+\n{qual}'
    return fastq.format(id=seq[0].decode('ascii'),
                        seq=seq[1].decode('ascii'),
                        qual=bytes(seq[2]).decode('ascii'))
                       
def writeToFifo(pairs,forward,reverse,fastq,logger):
    logger.info("Forward tmp file : %s" % forward)
    logger.info("Reverse tmp file : %s" % reverse)
    i=0
    
    if fastq:
        ffq=open("%s.odx/forward.fastq" % fastq,'w')
        rfq=open("%s.odx/reverse.fastq" % fastq,'w')
        
    with open(forward,'w',buffering=1) as f, \
         open(reverse,'w',buffering=1) as r:
        for p in pairs:
            i+=1
            if not i % 1000000:
                logger.info('%d sequence pairs written' % i)
            if len(p[0][1]) > 0 and len(p[1][1]) > 0:
                fread = formatFastq(p[0])
                rread = formatFastq(p[1])
                print(fread,file=f)
                print(rread,file=r)
                if fastq:
                    print(fread,file=ffq)
                    print(rread,file=rfq)
        
    if fastq:
        ffq.close()
        rfq.close()

def run(config):

    logger  = config['orgasm']['logger']
    output  = config['orgasm']['indexfilename']
    forward = config['index']['forward']
    reverse = config['index']['reverse']
    nopipe  = config['index']['nopipe']
    
    if config['index']['reformat']:
        reformatOldIndex(config)
    else:
        if forward is None:
            logger.error('No sequence file specified')
            sys.exit(1)
    
    if nopipe:
        logger.info("Indexing in no pipe mode")
    
    if reverse is None:
        if config['index']['mate']:
            mode="IMP"
        elif config['index']['single']:
            mode="SimPE"
        else:
            mode='IPE'
    else:
        if config['index']['mate']:
            mode="MP"
        elif config['index']['single']:
            logger.error('Two sequence files provided and --single specified')
            sys.exit(1)
        else:
            mode='PE'
  
    if not os.path.exists('%s.odx' % output ):
        os.makedirs('%s.odx' % output ) 
        
    command = ['orgasmi','-o','%s.odx/index'  % output]
          
    if config['index']['direct']:
        if mode!='PE':
            logger.error('Bypass filtering mode but no regular pair-end files used')
            sys.exit(1)
        else:
            if forward[-3:]=='.gz':
                logger.info('Forward file compressed by gzip')
                forward = ungzip(forward,nopipe)
                 
            if forward[-4:]=='.bz2':
                logger.info('Forward file compressed by bzip2')
                forward = unbzip(forward,nopipe)
         
                 
            if reverse[-3:]=='.gz':
                logger.info('Reverse file compressed by gzip')
                reverse = ungzip(reverse,nopipe)
          
            if reverse[-4:]=='.bz2':
                logger.info('Reverse file compressed by bzip2')
                reverse = unbzip(reverse,nopipe)
                
    else:
        filenames = [forward]
        if reverse is not None:
            filenames.append(reverse)
            
        forward = mktemp(prefix="forward-", dir=getTmpDir())
        atexit.register(postponerm(forward))
        
        if not nopipe:
            os.mkfifo(forward)
        
        reverse = mktemp(prefix="reverse-", dir=getTmpDir())
        atexit.register(postponerm(reverse))
        
        if not nopipe:
            os.mkfifo(reverse)
            
        setLogger(logger)

        pairs = readPairedEnd(filenames,
                              mode            = mode,
                              checkIds        = config['index']['checkpairs'],
                              cut             = config['index']['5trim'],
                              badQualityLimit = config['index']['badqual'],
                              qualityCut      = config['index']['qualtrim'],
                              length          = config['index']['length'],
                              lengthestimate  = config['index']['estimate'],
                              minlength       = config['index']['minlength'],
                              skip            = config['index']['skip'],
                              maxread         = config['index']['maxread'],
                              phix            = config['index']['phix'],
                              shift           = config['index']['shift'],
                              zip             = config['index']['zip']
                             )
        
        if config['index']['fastqdump']:
            fastq = output
        else:
            fastq = None
                                       
    command.append(forward)
    command.append(reverse)
    
    logger.info(" ".join(command))
    
    if not config['index']['direct']:
        if nopipe:
            logger.info("Writting transformed sequence files...")
            writeToFifo(pairs,forward,reverse,fastq,logger)
            logger.info("Done.")
            
            logger.info("Starting indexing...")
            logger.info(" ".join(command))
            os.system(" ".join(command)) 
        else:
            try:
                logger.info("Starting indexing...")
                process=Popen(command)
                writeToFifo(pairs,forward,reverse,fastq,logger)
                process.wait()
                logger.info('Done.')
            except BrokenPipeError:
                process.wait()
                logger.info('Done.')
                logger.warning("Maximum count of read indexed.")
                logger.warning("Indexing stopped but the index is usable.")
    else: 
        logger.info("Starting indexing...")
        logger.info(" ".join(command))
        os.system(" ".join(command)) 
    
        
    stats = getStats()
    
    if "paircount" in stats:
        logger.info("%9d reads pairs processed" % stats["paircount"])
    
    if "phix174" in stats:
        logger.info("%9d phix174 reads pairs sorted out" % stats["phix174"])
    
    if "SoftQualityTrim" in stats:
        logger.info("%9d reads pairs soft trimmed on a quality of %d" % (stats["SoftQualityTrim"],
                                                                         config['index']['badqual']))
    if "ACGTtrim" in stats:
        logger.info("%9d reads pairs clipped for not [A,C,G,T] bases" % stats["ACGTtrim"])
    
    if "bad_ids" in stats:
        logger.info("%9d reads pairs sorted out because of not corresponding ids" % stats["bad_ids"]) 
        
    
    
    r=Index('%s.odx/index'  % output)
            
    logger.info('Count of indexed reads: %d' % len(r))  

              



from ._asmbgraph cimport  AsmbGraph
from orgasm.graph._graphmultiedge cimport DiGraphMultiEdge
from orgasm.indexer._orgasm cimport Index
    


cdef class Assembler:
    cdef AsmbGraph _graph
    cdef Index _index
    cdef int _overlap
    cdef set extensionPoints 
    cdef int _startread
    cdef int _finalread
    cdef list _seeds
    cdef dict _annotations

    cpdef tuple readType(self,ids)

cpdef dict buildstem(Assembler assembler,
               int first,
               int last,
               bytes sequence,
               list path,
               bint circle)


# cdef class Stem(dict):
#     cdef int   _first
#     cdef int   _last
#     cdef int   _length
#     cdef bytes _sequence
#     cdef list  _path
#     cdef bint  _palindrome
#     cdef str   _label
#     cdef int   _weight
#     cdef bint  _circle
#     cdef int   _stemid
#  
#     cdef bint isPalindrome(self)
#     cdef str buildLabel(self)
#     cdef int getWeight(self, Assembler graph)  except 0
 
cdef class CompactAssembling(DiGraphMultiEdge):
    cdef dict _paths 
    cdef dict _stemid
    cdef bint _stemidOk
    cdef Assembler _assembler
    
    cdef int getStemid(self, dict stem) except 0
    cdef void setStemid(self)
    cpdef dict addStem(self, dict stem)
    
    
   

cdef class StemIterator:
    cdef dict       edgeName
    cdef Assembler  _assembler
    cdef AsmbGraph  _graph




#cython: language_level=3

cpdef str setRootConfigName(str rootname)
cpdef str getRootConfigName()

cdef dict buildDefaultConfiguration(str root_config_name,
                                    dict  config)

cpdef dict getConfiguration(str root_config_name=?,
                                    dict  config=?)


'''
Created on 10 mars 2015

@author: coissac
'''

try:
    from sphinx.setup_command import BuildDoc as ori_build_sphinx  # @UnresolvedImport
        
    class build_sphinx(ori_build_sphinx):
        '''Build Sphinx documentation in html, epub and man formats 
        '''
    
        description = __doc__
    
        def run(self):
            self.builder='html'
            self.finalize_options()
            ori_build_sphinx.run(self)
            self.builder='epub'
            self.finalize_options()
            ori_build_sphinx.run(self)
            self.builder='man'
            self.finalize_options()
            ori_build_sphinx.run(self)
except ImportError:
    pass  


# cython: language_level=3

cdef import from "stdlib.h":
    void* malloc(int size)  except NULL
    void* realloc(void* chunk,int size)  except NULL
    void free(void* chunk)
    
cdef import from "string.h":
    void bzero(void *s, size_t n)
    void memset(void* chunk,int car,int length)
    void memcpy(void* s1, void* s2, int n)
    
cdef struct AlignCell :
    int score
    int   path 
    
cdef struct AlignMatrix :
    AlignCell*  matrix
    int         msize
    int         vsize
    int         hsize
   


cdef AlignMatrix* allocateMatrix(int hsize, int vsize,AlignMatrix *matrix=?)

cdef void freeMatrix(AlignMatrix* matrix)

cdef void resetMatrix(AlignMatrix* matrix)

cdef struct alignPath:
    long length
    long buffsize
    long *path
    
cdef alignPath* allocatePath(long l1,long l2,alignPath* path=?)

cdef void reversePath(alignPath* path)


cdef void freePath(alignPath* path)



cdef class EndGapFree:
    cdef AlignMatrix* matrix

    cdef bytes horizontalSeq 
    cdef bytes verticalSeq
    
    cdef char* hSeq
    cdef char* vSeq
    cdef long  lhSeq
    cdef long  lvSeq
    cdef alignPath*     path

    cdef object alignment
    
    cdef bint sequenceChanged
    cdef bint matchScore(self,int h, int v)
    
    cdef int allocate(self) except -1
    cdef long doAlignment(self) except? 0
    cdef void reset(self)
    cdef inline int index(self, int x, int y)
    cdef void backtrack(self)
    cdef void clean(self)




from .fakereads cimport *
import traceback

cdef class FakeReads(dict):

    def __init__(self,size_t initid,size_t readsize=0):
        
        self._firstid = initid
        self._nextid  = initid
        self._readsize = readsize
        
        self._reverse = {}
        self._keys = {}
            
    cpdef bint isFake(self, int32_t id):
        cdef bint rep =  abs(id) >= self._firstid and abs(id) < self._nextid
        return rep
    
    cdef int _getid(self, bytes seq):
        cdef bytes cseq
        cdef object oid
        cdef char[50] buffer
        cdef char* pbuffer=buffer+1
        cdef char* nbuffer=buffer
        cdef int id
        
        seq = seq.upper()
        
        oid = self.get(seq,None)

        if  oid is None:
            
            assert PyBytes_GET_SIZE(seq)==self._readsize
            
            #print 'Create a new fake sequence'
            #traceback.print_stack()
            
            id = self._nextid
            self._nextid+=1
            
            self[seq]=PyInt_FromLong(id)
            
            cseq = reverseComplement(seq)
            self[cseq]=PyInt_FromLong(-id)
            
            snprintf(pbuffer,49,b"%d",id)
            nbuffer[0]=b'-'
            self._reverse[pbuffer]=seq
            self._reverse[nbuffer]=cseq
        else:
            id = PyInt_AsLong(oid)
            
        return id
            
            
    cpdef tuple getReadIds(self, bytes seq):
        cdef int id = self._getid(seq)
    
        return (id,0,set([id]))

    cpdef tuple getIds(self, int32_t id):
        assert self.isFake(id)
        return (abs(id),0,set([abs(id)]))
    
    cpdef bytes getRead(self, int32_t readid, int32_t begin, int32_t length):
        cdef char[50] buffer
        cdef int id
        cdef bytes sequence
        cdef char* pseq

        assert self.isFake(readid)

        snprintf(buffer,50,b"%d",readid)
        sequence = self._reverse[buffer]
        pseq = <char*>sequence + begin 
        return PyBytes_FromStringAndSize(pseq,length)
     
    cpdef int len(self):
        return PyDict_Size(self)
        
    property firstid:

        "A doc string can go here."

        def __get__(self):
            return self._firstid

    property lastid:

        "A doc string can go here."

        def __get__(self):
            return self._nextid

            
            


"""
.. module:: orgasm.assembler
   :platform: Unix
   :synopsis: The :py:mod:`orgasm.assembler` python package provide the :class:`Assembler` class .

The :py:mod:`orgasm.assembler` package
======================================

:author:  Eric Coissac
:contact: eric.coissac@inria.fr

The :py:mod:`orgasm.assembler` python package provide the :class:`Assembler` class 
which manage the assembling process.
"""

from ._assembler import Assembler       # @UnresolvedImport
from ._assembler import buildstem       # @UnresolvedImport
from ._tango import tango,getusedreads,resetusedreads  # @UnresolvedImport



'''
Created on 28 sept. 2014

@author: coissac
'''


from orgasm import getIndex, getSeeds,getOutput
from orgasm.tango import  estimateFragmentLength,\
    genesincontig, scaffold,  restoreGraph


__title__="Recompact the assembling graph"
 
default_config = { 'seeds' : None
                 }


def addOptions(parser):
    parser.add_argument(dest='orgasm:indexfilename',  metavar='index', 
                        help='index root filename (produced by the oa index command)')
    
    parser.add_argument(dest='orgasm:outputfilename',     metavar='output', 
                                                          nargs='?', 
                                                          default=None,
                        help='output prefix' )
    
    
    
    parser.add_argument('--back',             dest='orgasm:back', 
                                              type=int, 
                                              action='store', 
                                              default=None, 
                        help='the number of bases taken at the end of '
                             'contigs to jump with pared-ends [default: <estimated>]')
    


def run(config):
    
    logger=config['orgasm']['logger']
    output = getOutput(config)
    
    r = getIndex(config)
    coverage,x,newprobes = getSeeds(r,config)  
    
    asm = restoreGraph(output+'.oax',r,x)
    meanlength,sdlength = estimateFragmentLength(asm)

    if config['orgasm']['back'] is not None:
        back = config['orgasm']['back']
    elif config['orgasm']['back'] is None and meanlength is not None:
        back = int(meanlength + 4 * sdlength)
        if back > 500:
            back=500
    else:
        back = 300
        
    if meanlength is not None:
        logger.info("Fragment length estimated : %f pb (sd: %f)" % (meanlength,sdlength))

        
    cg = asm.compactAssembling(verbose=False)
    logger.info("Scaffold the assembly")
    scaffold(asm,cg,minlink=5,back=int(back),addConnectedLink=False,
                 logger=logger)
    genesincontig(cg,r,x)
    with open(output+'.gml','w') as gmlfile:
        print(cg.gml(),file=gmlfile)



#cython: language_level=3
from cpython.array cimport array
from orgasm.files.universalopener import uopen


def readFasta(filename,int quality=40,int shift=33):
    
    cdef int   seqid = 0
    cdef int   cut_slash
    cdef int   cut_space
    
    cdef bytes seq
    cdef list  seqlines = []
    
    if isinstance(filename, str):
        filename = uopen(filename)
    
    try:        
        line  = next(filename) 
        bline = bytes(line,
                      encoding='ascii'
                     )

        while(bline[0]!=62):   # ord(b'>')
            bline = bytes(next(filename),
                          encoding='ascii'
                         )

        while(1):
                            
            cut_space = bline.find(b' ')
            cut_slash = bline.find(b'/')
            if cut_space >= 0 :
                if cut_space < cut_slash or cut_slash < 0:
                    sid = bline[1:cut_space]
                else:
                    sid = bline[1:cut_slash]
            elif cut_slash >= 0:
                sid = bline[1:cut_slash]
            else:
                sid = bline.strip()
                
            bline = bytes(next(filename),
                          encoding='ascii'
                         )
                
            while (bline[0]!=62):   # ord(b'>')
                seqlines.append(bline.strip().upper())
                
                bline = bytes(next(filename),
                              encoding='ascii'
                             )
                
            seq = b''.join(seqlines)
            yield (sid,
                   seq,
                   array('B',[quality+shift] * len(seq)))
            seqlines=[]
                
    except StopIteration:
        pass
    
    seq = b''.join(seqlines)
    yield (sid,
           seq,
           array('B',[quality+shift] * len(seq)))

                    



#cython: language_level=3

'''
Created on 28 mars 2016

@author: coissac
'''

import zipfile
import bz2
import gzip

import io

cdef class MagicKeyFile:
    def __init__(self,stream,length=2):
        
        binary=stream
        self.stream = stream
        self.stream_mode = None 
        if hasattr(stream, "mode"):
            self.stream_mode = stream.mode
            if (not 'b' in stream.mode and 
                hasattr(stream, "buffer") and 
                'b' in stream.buffer.mode):
                binary=stream.buffer

        if (self.stream_mode is None and 
            not (hasattr(stream, 'headers') and 
                 hasattr(stream.headers, "keys") and
                 'Content-type' in stream.headers)):
            raise TypeError("stream does not present the good interface")
        
        self.binary=binary
        self.key=binary.read(length)
        self.keylength=length
        self.pos=0
        
    cpdef bytes read(self,int size=-1):
        cdef bytes r
        
        if self.pos < self.keylength:
            if size > (self.keylength - self.pos):
                size = size - self.keylength + self.pos
                r = self.key[self.pos:] + self.binary.read(size)
                self.pos=self.keylength + 1
            elif size >=0 :
                r = self.key[self.pos:(self.pos+size)]
                self.pos+=size
            else:
                r = self.key[self.pos:] + self.binary.read(size)
                self.pos=self.keylength + 1
        else:
            r = self.binary.read(size)
            
        return r

    cpdef bytes read1(self,int size=-1):
        return self.read(size)

    cpdef int tell(self):
        cdef int p
        
        if self.pos < self.keylength:
            p = self.pos 
        else:
            p = self.binary.tell()
        
        return p
    
    def __getattr__(self,name):
        return getattr(self.binary, name)
        

    
cdef class CompressedFile:
    
    
    def __init__(self,stream):
        cdef int            keylength
        cdef MagicKeyFile   magic
        cdef str            compressor
        cdef bytes          k 
        cdef object         c
        
        cdef dict compress = { 'zip' : (b'\x50\x4b\x03\x04',zipfile.ZipFile),
                               'bz2' : (b'\x42\x5a\x68',bz2.BZ2File),
                               'gz'  : (b'\x1f\x8b\x08',gzip.open)
                   }
        
        keylength = max([len(x[0]) for x in compress.values()])
        magic=MagicKeyFile(stream,keylength)
        
        self.accessor = None
        
        for compressor in compress:
            k,c = compress[compressor]
            if magic.key.startswith(k):
                self.accessor = c(magic)
        
        if self.accessor is None:
            self.accessor = magic
        
        if ((hasattr(stream, 'headers') and 
             hasattr(stream.headers, "keys") and
             'Content-type' in stream.headers and
             stream.headers['Content-type'].startswith('text/')) or
            'b' not in magic.stream_mode):
            self.accessor = io.TextIOWrapper(self.accessor)
        
    
    def __getattr__(self,name):
        return getattr(self.accessor, name)
    
    def __iter__(self):
        for x in self.accessor:
            yield x 


# cython: language_level=3

from ._fibointheap cimport *

import sys

cdef class FibonacciIntHeap:

    def __dealloc__(self):
        fiboTreeExit(&self._treedat)
        for i in range(self._arenacount):
            free(self._arenas[i])
        free(self._arenas)
                
    cdef void _realloc(self):
        cdef IntData**  arenas
        cdef IntData*   newarena
        cdef size_t i
        
        if self._arenacount==0:
            arenas = <IntData**>malloc(10 * sizeof(IntData*))
        elif self._arenacount % 10 == 0:
            arenas = <IntData**>realloc(<void*>self._arenas,
                                        (self._arenacount+10) * sizeof(IntData*))
        else:
            arenas = self._arenas
            
        if arenas==NULL:
            raise SystemError
     
        self._arenas=arenas
            
        newarena = <IntData*>malloc(sizeof(IntData) * self._chunksize)
    
        if newarena==NULL:
            raise SystemError

        arenas[self._arenacount]=newarena
        self._arenacount+=1
        
        for i in range(self._chunksize-1):
            (<IntData**>(newarena+i))[0]=newarena+i+1
            
        (<IntData**>(newarena+self._chunksize-1))[0]=NULL
        
        if self._arenacount>1:
            (<IntData**>(self._arenas[self._arenacount-1]+self._chunksize-1))[0]=newarena
            
        self._last = newarena
        
    cdef void _setNode(self,IntData* node,int32_t data,size_t priority):
        
        fiboTreeDel(&self._treedat,<FiboNode*>node)
        node.priority = priority 
        node.data=data
        fiboTreeAdd(&self._treedat,<FiboNode*>node)
        

    cpdef size_t push(self,int32_t data,size_t priority) except * :
        cdef IntData* d = self._last
        cdef object odata = PyInt_FromLong(data)
        
        assert odata not in self._nodes,'you cannot insert two times the same key'
                
        self._nodes[odata]=PyInt_FromSsize_t(<size_t>self._last)
        self._last= (<IntData**>self._last)[0]
        
        if self._last == NULL:
            self._realloc()
        
        d.priority = priority
        d.data     = data 
        
        fiboTreeAdd(&self._treedat, &d.node)
       
        self._size+=1
                        
        return priority
               
    cpdef tuple pop(self):
        cdef FiboNode *node 
        cdef IntData     *d
        cdef size_t   n
        
        assert self._size > 0,"You cannot pop from an empty heap"
        
        node = fiboTreeMin(&self._treedat)
        d = <IntData*>node
        data = d.data 
        fiboTreeDel(&self._treedat,node)
        self._size-=1
                       
        (<IntData**>d)[0]=self._last
        self._last = d
        
        del self._nodes[PyInt_FromLong(d.data)]
        
        return d.data,d.priority
         
    cpdef int32_t popInt(self):
        cdef FiboNode *node 
        cdef IntData     *d
        cdef size_t   n
        
        assert self._size > 0,"You cannot pop from an empty heap"
        
        node = fiboTreeMin(&self._treedat)
        d = <IntData*>node
        data = d.data 
        fiboTreeDel(&self._treedat,node)
        self._size-=1
                       
        (<IntData**>d)[0]=self._last
        self._last = d
        
        del self._nodes[PyInt_FromLong(d.data)]
                
        return d.data 

    def __len__(self):
        return self._size
    
    def __iter__(self):
        
        while(self._size > 0):
            yield self.popInt()
            
    cpdef size_t getPriority(self, int32_t key)  except *:
        cdef IntData *d
        
        d = <IntData*>PyInt_AsSsize_t(self._nodes[PyInt_FromLong(key)])
        return d.priority
    
    def __getitem__(self,key):
        return self.getPriority(key)

    cpdef size_t setPriority(self, int32_t key, size_t priority) except *:
        cdef IntData *d = <IntData*>PyInt_AsSsize_t(self._nodes[PyInt_FromLong(key)])
        self._setNode(d,key,priority)
        return priority
    
    def __setitem__(self,int32_t key, size_t priority):
        try:
            self.setPriority(key,priority)
        except KeyError:
            self.push(key,priority)
            
    def __contains__(self,int32_t key):
        cdef object k = PyInt_FromLong(key)
        return k in self._nodes
    
        



#cython: language_level=3

cdef class MagicKeyFile:
    cdef object stream
    cdef str    stream_mode
    cdef object binary
    cdef bytes  key 
    cdef int    keylength 
    cdef int    pos
    
    cpdef bytes read(self,int size=?)
    cpdef bytes read1(self,int size=?)
    cpdef int tell(self)

    
cdef class CompressedFile:
    cdef object accessor
    


'''
Created on 10 mars 2015

@author: coissac
'''

import os.path

from distutils.command.sdist import sdist as orig_sdist
from distutils import dir_util

class sdist(orig_sdist):

    def make_distribution(self):
        """Create the source distribution(s).  First, we create the release
        tree with 'make_release_tree()'; then, we create all required
        archive files (according to 'self.formats') from the release tree.
        Finally, we clean up by blowing away the release tree (unless
        'self.keep_temp' is true).  The list of archive files created is
        stored so it can be retrieved later by 'get_archive_files()'.
        """
        # Don't warn about missing meta-data here -- should be (and is!)
        # done elsewhere.
        base_dir = self.distribution.get_fullname()
        base_name = os.path.join(self.dist_dir,base_dir)

        self.make_release_tree(os.path.join('tmp',base_dir), self.filelist.files)
        archive_files = []              # remember names of files we create
        # tar archive must be created last to avoid overwrite and remove
        if 'tar' in self.formats:
            self.formats.append(self.formats.pop(self.formats.index('tar')))

        for fmt in self.formats:
            file = self.make_archive(base_name, fmt, root_dir='tmp',base_dir=base_dir,
                                     owner=self.owner, group=self.group)
            archive_files.append(file)
            self.distribution.dist_files.append(('sdist', '', file))

        self.archive_files = archive_files

        if not self.keep_temp:
            dir_util.remove_tree(os.path.join('tmp',base_dir), dry_run=self.dry_run)



'''
Created on 28 sept. 2014

@author: coissac
'''
import sys

from orgasm import getOutput,getIndex, getSeeds
from orgasm.tango import restoreGraph, estimateFragmentLength, genesincontig,\
    scaffold, path2fasta, unfoldmarker, parseFocedScaffold

__title__="Assembling graph unfolder for the nuclear rDNA complex"

default_config = {
                 }

def addOptions(parser):
    parser.add_argument(dest='orgasm:indexfilename',  metavar='index', 
                        help='index root filename (produced by the oa index command)')
    
    parser.add_argument(dest='orgasm:outputfilename',     metavar='output', 
                                                          nargs='?', 
                                                          default=None,
                        help='output prefix' )
    
    
    parser.add_argument('--force-scaffold',   dest='unfold:fscaffold', 
                                              action='append', 
                                              default=None, 
                        help='Force circular unfolding')

    parser.add_argument('--back',             dest='orgasm:back', 
                                              type=int, 
                                              action='store', 
                                              default=None, 
                        help='the number of bases taken at the end of '
                             'contigs to jump with pared-ends [default: <estimated>]')
    
    parser.add_argument('--set-tag','-S',     dest ='unfold:tags', 
                                              metavar='tag', 
                                              action='append',
                                              default=[], 
                                              type=str, 
                        help='Allows to add a tag in the OBITools format '
                             'to the header of the fasta sequences')


def selectGoodComponent(cg):
    
    def geneincc(cc):
        return any(cg.getEdgeAttr(*e)['ingene']>0 for e in cc)
    
    cc = list(cg.connectedComponentIterator())
    goodcc=[]
    for cc1 in cc:
        ccp = set(-i for i in cc1)
        if ccp not in goodcc:
            goodcc.append(cc1)
      
    gcc = [list(cg.edgeIterator(nodePredicate=lambda n:n in c,
                                edgePredicate=lambda e: 'stemid' in cg.getEdgeAttr(*e))) 
           for c in goodcc] 
    
    gcc = [cc for cc in gcc if geneincc(cc)]

    return gcc
    


def run(config):

    logger=config['orgasm']['logger']
    output = getOutput(config)

    forcedscaffold = parseFocedScaffold(config['unfold']['fscaffold'])

    r = getIndex(config)
    coverage,x,newprobes = getSeeds(r,config)  # @UnusedVariable
    
    asm = restoreGraph(output+'.oax',r,x)

    logger.info("Evaluate fragment length")
    
    meanlength,sdlength = estimateFragmentLength(asm)
    
    if meanlength is not None:
        logger.info("Fragment length estimated : %f pb (sd: %f)" % (meanlength,sdlength))

    if config['orgasm']['back'] is not None:
        back = config['orgasm']['back']
    elif config['orgasm']['back'] is None and meanlength is not None:
        back = int(meanlength + 4 * sdlength)
        if back > 500:
            back=500
    else:
        back = 300
        
    logger.info("Evaluate pair-end constraints")
    
    cg = asm.compactAssembling(verbose=False)
    
    genesincontig(cg,r,x)

    scaffold(asm,
             cg,
             minlink=5,
             back=int(back),
             addConnectedLink=False,
             forcedLink=forcedscaffold,
             logger=logger)
     
    fastaout = sys.stdout

    pathout  = open(output+".path","w")
    
    logger.info("Select the good connected components")
    gcc = selectGoodComponent(cg)
    
    logger.info("Print the result as a fasta file")
    
    c=1
    seqid = config['orgasm']['outputfilename'].split('/')[-1]

    for seeds in gcc:
        path = unfoldmarker(cg,seeds=seeds)
                
        fa = path2fasta(asm,cg,path,
             identifier="%s_%d" % (seqid,c),
             back=back,
             minlink=config['orgasm']['minlink'],
             logger=logger,
             tags=config['unfold']['tags'])
        
        print(fa, file=fastaout)
        print(" ".join([str(x) for x in path]),file=pathout)
        c+=1
        
    print(cg.gml(),file=open(output +'.path.gml','w'))

              



'''
Created on 20 oct. 2012

@author: coissac
'''

from os import path
import os.path
import glob
import sys

# try:
#     from setuptools.extension import Extension
# except ImportError:
#     from distutils.extension import Extension

from distutils.extension import Extension

from obidistutils.serenity.checkpackage import  install_requirements,\
                                                check_requirements, \
                                                RequirementError
                                                
from obidistutils.serenity.rerun import enforce_good_python
from obidistutils.serenity.rerun import rerun_with_anothe_python
from distutils import log

from obidistutils.dist import Distribution
from obidistutils.serenity import is_serenity
 
 
def findPackage(root,base=None):
    modules=[]
    if base is None:
        base=[]
    for module in (path.basename(path.dirname(x)) 
                   for x in glob.glob(path.join(root,'*','__init__.py'))):
        modules.append('.'.join(base+[module]))
        modules.extend(findPackage(path.join(root,module),base+[module]))
    return modules
    
def findCython(root,base=None,pyrexs=None):
    setupdir = os.path.dirname(sys.argv[0])
    pyrexs=[]

    if base is None:
        base=[]
    for module in (path.basename(path.dirname(x)) 
                   for x in glob.glob(path.join(root,'*','__init__.py'))):
                       
                
        for pyrex in glob.glob(path.join(root,module,'*.pyx')):
            pyrexs.append(Extension('.'.join(base+[module,path.splitext(path.basename(pyrex))[0]]),
                                    [pyrex]
                                    )
                          )
            try:
                cfiles = os.path.splitext(pyrex)[0]+".cfiles"
                cfilesdir = os.path.dirname(cfiles)
                cfiles = open(cfiles)
                cfiles = [os.path.relpath(os.path.join(cfilesdir,y),setupdir).strip() 
                          if y[0] !='@' else y.strip()
                          for y in cfiles]
                
                log.info("Cython module : %s",cfiles)   
                incdir = set(os.path.dirname(x) for x in cfiles if x[-2:]==".h")
                cfiles = [x for x in cfiles if x[-2:]==".c"]                
                pyrexs[-1].sources.extend(cfiles)
                pyrexs[-1].include_dirs.extend(incdir)
                pyrexs[-1].extra_compile_args.extend(['-msse2',
                                                      '-Wno-unused-function',
                                                      '-Wmissing-braces',
                                                      '-Wchar-subscripts'])
                
            except IOError:
                pass
            
        pyrexs.extend(findCython(path.join(root,module),base+[module]))
    return pyrexs
    

def rootname(x):
    return os.path.splitext(x.sources[0])[0]


def prepare_commands():
    from obidistutils.command.build import build
    from obidistutils.command.littlebigman import littlebigman
#    from obidistutils.command.serenity import serenity
    from obidistutils.command.build_cexe import build_cexe
    from obidistutils.command.build_ext import build_ext
    from obidistutils.command.build_ctools import build_ctools
    from obidistutils.command.build_files import build_files
    from obidistutils.command.build_scripts import build_scripts
    from obidistutils.command.install_scripts import install_scripts
    from obidistutils.command.install_sphinx import install_sphinx
    from obidistutils.command.install import install
    from obidistutils.command.pidname import pidname
    from obidistutils.command.sdist import sdist
    
        
    
    COMMANDS = {'build':build,
#                'serenity':serenity,
                'littlebigman':littlebigman,
                'pidname':pidname,
                'build_ctools':build_ctools, 
                'build_files':build_files,
                'build_cexe':build_cexe, 
                'build_ext': build_ext,
                'build_scripts':build_scripts, 
                'install_scripts':install_scripts,
                'install_sphinx':install_sphinx,
                'install':install,
                'sdist':sdist}
    
#     try:
#         from setuptools.commands import egg_info
#         COMMANDS['egg_info']=egg_info
#     except ImportError:
#         pass
    try:
        from obidistutils.command.build_sphinx import build_sphinx
        COMMANDS['build_sphinx']=build_sphinx
    except ImportError:
        pass

    return COMMANDS


CTOOLS =[]
CEXES  =[]
FILES  =[]

def setup(**attrs):
    
    log.set_threshold(log.INFO)
    
    minversion      = attrs.get("pythonmin",'3.4')
    maxversion      = attrs.get('pythonmax',None)    
    fork            = attrs.get('fork',False)
    requirementfile = attrs.get('requirements','requirements.txt')

    try:
        del attrs['pythonmin']
    except KeyError:
        pass

    try:
        del attrs['pythonmax']
    except KeyError:
        pass

    try:
        del attrs['fork']
    except KeyError:
        pass

    try:
        del attrs['requirements']
    except KeyError:
        pass
    
    if is_serenity():
    
        
        enforce_good_python(minversion, maxversion, fork)
        
        if (install_requirements(requirementfile)):
            rerun_with_anothe_python(sys.executable,minversion,maxversion,fork)
        
    
        try:
            check_requirements(requirementfile)
        except RequirementError as e :
            log.error(e)                                   
            sys.exit(1)
        
    if 'distclass' not in attrs:
        attrs['distclass']=Distribution

    if 'python_src' not in attrs:
        SRC = 'python'
    else:
        SRC = attrs['python_src']
        del(attrs['python_src'])
    
    if 'scripts' not in attrs:
        attrs['scripts'] = glob.glob('%s/*.py' % SRC)

    if 'package_dir' not in attrs:
        attrs['package_dir'] = {'': SRC}

    if 'packages' not in attrs:
        attrs['packages'] = findPackage(SRC)
    
    if 'cmdclass' not in attrs:
        attrs['cmdclass'] = prepare_commands()

    if 'ctools' not in attrs:
        attrs['ctools'] = CTOOLS
    
    if 'executables' not in attrs:
        attrs['executables'] = CEXES
        
    if 'files' not in attrs:
        attrs['files'] = FILES
        
    if 'sse' not in attrs:
        attrs['sse']=None
        
    if 'serenity' not in attrs:
        attrs['serenity']=False
    
    EXTENTION=findCython(SRC)
    
    if 'ext_modules' not in attrs:
        attrs['ext_modules'] = EXTENTION
        
#     try:
#         from setuptools.core import setup as ori_setup
#     except ImportError:
#         from distutils.core import setup as ori_setup

    from distutils.core import setup as ori_setup
    
    ori_setup(**attrs)



# cython: language_level=3

from cpython.dict cimport *
from cpython.int cimport PyInt_FromLong, PyInt_AsLong
from cpython.bytes cimport PyBytes_FromStringAndSize, PyBytes_GET_SIZE
from libc.stdio cimport snprintf
from orgasm.utils.dna cimport *

cdef extern from "inttypes.h":
    ctypedef int int32_t  
    ctypedef unsigned int uint32_t

cdef extern from "stdlib.h":      
    ctypedef unsigned int size_t  
    int32_t abs(int32_t i)
    
cdef extern from "math.h":
    double log10(double x)
    double ceil(double x)

cdef class FakeReads(dict):

    cdef dict   _reverse
    cdef dict   _keys
    cdef size_t _firstid
    cdef size_t _nextid
    cdef size_t _readsize
    
    cdef int _getid(self, bytes seq)
    cpdef bint isFake(self, int32_t id)
    cpdef tuple getReadIds(self, bytes seq)
    cpdef tuple getIds(self, int32_t id)    
    cpdef bytes getRead(self, int32_t readid, int32_t begin, int32_t length)
    cpdef int len(self)



#cython: language_level=3

'''
Created on 27 mars 2016

@author: coissac
'''

import logging
import sys

cpdef getLogger(dict config):
    '''
    Returns the logger as defined by the command line option
    or by the config file
    :param config:
    '''
    
    root = config["__root_config__"]
 
    level  = config[root]['loglevel'] 
    logfile= config[root]['log'] 
    
    rootlogger   = logging.getLogger()
    logFormatter = logging.Formatter("%(asctime)s [%(levelname)-5.5s]  %(message)s")

    stderrHandler = logging.StreamHandler(sys.stderr)
    stderrHandler.setFormatter(logFormatter)

    rootlogger.addHandler(stderrHandler)
    
    if logfile:
        fileHandler = logging.FileHandler(logfile)
        fileHandler.setFormatter(logFormatter)
        rootlogger.addHandler(fileHandler)
    
    try:
        loglevel = getattr(logging, level) 
    except:
        loglevel = logging.INFO
        
    rootlogger.setLevel(loglevel)
    
    config[root]['logger']=rootlogger
    
    return rootlogger



# cython: language_level=3

cdef extern from "inttypes.h":
    ctypedef int int32_t  
    ctypedef int int64_t  
    ctypedef unsigned int uint32_t
    ctypedef unsigned int uint64_t


cdef class BitVector:

    cdef size_t         _size
    cdef size_t         _wsize
    cdef size_t         _vcount
    cdef size_t         *_count
    cdef uint32_t       *_data

    cpdef size_t set(self, size_t v, size_t pos)
    cpdef size_t unset(self, size_t v, size_t pos)
    cpdef size_t get(self,size_t v)
    cpdef clear(self)
    cdef size_t* counter(self)
    
        




'''
Created on 28 sept. 2014

@author: coissac
'''


__title__="Substract an assembly from an read index"

from orgasm import getIndex,getSeeds
from orgasm.backtranslate.fasta import fasta
from orgasm.tango import restoreGraph

from subprocess import Popen
from tempfile   import mkdtemp
from tempfile   import mktemp
from shutil     import rmtree

import atexit
import os.path
import sys
import math


default_config = { "assembly" : None,
                   'sequence' : None,
                   "circular" : False
                 }

tmpdir = []

def addOptions(parser):
    parser.add_argument(dest='orgasm:indexfilename',  metavar='index', 
                        help='index root filename (produced by the oa index command)')
    
    parser.add_argument(dest='substract:newindex',  metavar='newindex', 
                        help='name of the new read index')
    
    
    parser.add_argument('--assembly',"-a",    dest ='substract:assembly', 
                                              metavar='assembly', 
                                              action='store',
                                              default=None, 
                                              type=str, 
                        help='The result of an assembly to substract from the read index')
    
    parser.add_argument('--sequence',"-s",    dest ='substract:sequence', 
                                              metavar='sequence', 
                                              action='store',
                                              default=None, 
                                              type=str, 
                        help='A fasta file containing sequences to substract from the read index')

    parser.add_argument("--circular","-c",
                        dest="substract:circular",
                        action='store_true',
                        default=None,
                        help='The fasta sequence file contains circular sequences'
                       )

def formatFastq(seq):
    
    fastq = '@{id:0>7}\n{seq}\n+\n{qual}'
    return fastq.format(id=seq[0],
                        seq=seq[1].decode('ascii'),
                        qual='0'*len(seq[1]))

def postponerm(filename):
    def trigger():
        print("Deleting tmp file : %s" % filename,file=sys.stderr)
        try:
            os.unlink(filename)
        except:
            pass
        
    return trigger

def postponermdir(filename):
    def trigger():
        print("Deleting tmp directory : %s" % filename,file=sys.stderr)
        try:
            rmtree(filename)
        except:
            pass
        
    return trigger

def getTmpDir():
    global tmpdir

    if not tmpdir:
        tmpdir.append(mkdtemp())
        atexit.register(postponermdir(tmpdir[0]))
        
    return tmpdir[0]


def writeToFifo(index,excluded,forward,reverse,logger):
    logger.info("Forward tmp file : %s" % forward)
    logger.info("Reverse tmp file : %s" % reverse)

    maxreadid = len(index)+1
    readlength= index.getReadSize()
    step=10 ** math.floor(math.log10(maxreadid))

    with open(forward,'w') as f, \
         open(reverse,'w') as r:
        
        j=0
        printed=set()
        logger.info('%d sequence pairs written' % 0)
        
        for i in range(1,maxreadid):
            ri=index.getPairedRead(i)

            if (not (i in printed  or abs(ri) in printed or
                     i in excluded or abs(ri) in excluded)):    
                    
                forward = index.getRead(i,0,readlength)
                reverse = index.getRead(ri,0,readlength)
                
                printed.add(i)
                printed.add(abs(ri))
                
#                logger.info("coucou1")
                print(formatFastq((i,forward)), file=f)
#                logger.info("coucou2")
                print(formatFastq((ri,reverse)),file=r)
                logger.info("coucou3")

                if not j % step:
                    logger.info('%d sequence pairs written' % j)
                j+=1
    
        

def run(config):
    
    logger=config['orgasm']['logger']
    progress = config['orgasm']['progress']
    circular = config['substract']['circular']
    output   = config['substract']['newindex']
    
    r = getIndex(config)
    maxreadid = len(r)
    readlength= r.getReadSize()

    if config['substract']['assembly'] is not None:
        assembly="%s.oas/assembling.oax"  % config['substract']['assembly']

        if not os.path.exists(assembly):
            raise RuntimeError("Cannot load the assembly %s" % config['substract']['assembly'])
        
        ecoverage,x,newprobes = getSeeds(r,config)
        asm = restoreGraph(assembly,r,x)
        
        graph = asm.graph
        readids = set(abs(x) for x in graph.nodeIterator() if not r.fakes.isFake(x))

        print(len(asm))

                      
    elif config['substract']['sequence'] is not None:
        
        sequences = fasta(config['substract']['sequence'])
        
        readids = set()
        for seq in sequences.values():
            if circular:
                seq = seq+seq[0:readlength]
            for p in range(len(seq)-readlength):
                read   = seq[p:(p+readlength)]
                rid = r.getReadIds(read)[0]
                if not r.fakes.isFake(rid):
                    readids.add(abs(rid))

    allreads=set()
    
    for rid in readids:
        allreads|=set(abs(x) for x in r.getIds(rid)[2])
        paired = set(r.normalizedPairedEndsReads(rid)[1]) | set(r.normalizedPairedEndsReads(-rid)[1])
        for prid in paired:
            allreads|=set(abs(x) for x in r.getIds(prid)[2])
            
    tmpdir = getTmpDir()
    forward = mktemp(prefix="forward-", dir=tmpdir)
    atexit.register(postponerm(forward))

    reverse = mktemp(prefix="reverse-", dir=tmpdir)
    atexit.register(postponerm(reverse))
        
    os.mkfifo(forward)
    os.mkfifo(reverse)
    
    if not os.path.exists('%s.odx' % output ):
        os.makedirs('%s.odx' % output ) 

    command = ['orgasmi','-o','%s.odx/index'  % output]
    
    command.append(forward)
    command.append(reverse)

    logger.info(' '.join(command))

    try:
        logger.info("Starting indexing...")
        process=Popen(command)
        writeToFifo(r,allreads,forward,reverse,logger)
        process.wait()
        logger.info('Done.')
    except BrokenPipeError:
        process.wait()
        logger.info('Done.')
        logger.warning("Maximum count of read indexed.")
        logger.warning("Indexing stopped but the index is usable.")



# cython: language_level=3

from cpython cimport *
from .fibo cimport *
from libc.stdlib  cimport malloc,realloc,free

cdef struct Data:
    FiboNode            node                    
    int                 priority
    PyObject*           data                      

cdef int cmpFunc (const_FiboNode_ptr_const node1, const_FiboNode_ptr_const node2)

cdef class FibonacciHeap:

    cdef FiboTree _treedat
    cdef Data*    _arena
    cdef size_t   _arenasize
    cdef size_t   _last 
    cdef size_t   _chunksize
    cdef size_t   _size

    cdef void _realloc(self)
    cpdef push(self,int priority,object data)
    cpdef tuple min(self)
    cpdef object minItem(self)
    cpdef tuple pop(self)
    cpdef object popItem(self)




#cython: language_level=3

#from ..utils cimport str2bytes

cdef extern from "stdio.h":
    struct FILE
    int fprintf(FILE *stream, char *format, ...)
    int fputs(char *string, FILE *stream)
    FILE* stderr
    ctypedef unsigned int off_t "unsigned long long"
    
cdef extern from "unistd.h":
    int fsync(int fd);

cdef extern from "time.h":
    struct tm :
        int tm_yday 
        int tm_hour
        int tm_min
        int tm_sec

    enum: CLOCKS_PER_SEC
    
    ctypedef int time_t
    ctypedef int clock_t
    ctypedef int suseconds_t
    
    struct timeval:
        time_t      tv_sec     #  seconds */
        suseconds_t tv_usec    #  microseconds */


    struct timezone :
        int tz_minuteswest;    # minutes west of Greenwich
        int tz_dsttime;        # type of DST correction 

        
    int gettimeofday(timeval *tv, timezone *tz)


    tm *gmtime_r(time_t *clock, tm *result)
    time_t time(time_t *tloc)
    clock_t clock()

cdef class ProgressCounter:
    cdef off_t   digits
    cdef off_t   logeach
    cdef clock_t starttime
    cdef clock_t lasttime
    cdef clock_t tickcount
    cdef int freq
    cdef int cycle
    cdef int lastlog
    cdef bint ontty
    cdef int fd
    
    cdef bytes _head
    cdef char *chead
    cdef bytes _unit
    cdef char *cunit
    
    cdef object logger
    
    cdef clock_t clock(self)



'''
Created on 20 oct. 2012

@author: coissac
'''

# try:
#     from setuptools.dist import Distribution as ori_Distribution
# except ImportError:
#     from distutils.dist import Distribution as ori_Distribution

from distutils.dist import Distribution as ori_Distribution

class Distribution(ori_Distribution):
    
    def __init__(self,attrs=None):
        self.executables = None
        self.ctools = None
        self.files = None
        self.build_cexe  = None
        self.deprecated_scripts = None
        self.zip_safe=False
        self.sse = None
        self.serenity=attrs['serenity']
        
        ori_Distribution.__init__(self, attrs)
        
        self.global_options.insert(0,('serenity', None, "install or build the package in a python virtualenv "
                                                       "without polluting the installed python and with many "
                                                       "checks during the installation process"
                                     ))
        self.global_options.insert(0,('virtualenv', None, "if the installation is done using the serenity mode "
                                                        "this option allows for specifying the virtualenv name. "
                                                        "By default the name is PACKAGE-VERSION"
                                     ))
        
    def run_commands(self):
        """Run each command that was seen on the setup script command line.
        Uses the list of commands found and cache of command objects
        created by 'get_command_obj()'.
        """
#        self.run_command('littlebigman')
        ori_Distribution.run_commands(self)
        
        
    def has_executables(self):
        return self.executables is not None and self.executables
        
    def has_ctools(self):
        return self.ctools is not None and self.ctools
        
    def has_files(self):
        return self.files is not None and self.files
        
    def has_deprecated_scripts(self):
        return self.deprecated_scripts is not None and self.deprecated_scripts

    


# cython: language_level=3

from ._fibominintheap cimport *
from ._fibomaxintheap cimport *

from libc.stdio cimport snprintf
from libc.stdlib cimport atol

cdef extern from "inttypes.h":
    cdef enum:
        INT32_MAX
        INT32_MIN
        
cdef class NodeIterator:

    cdef object   _nodepredicate
    cdef object   _nodeiterator
    cdef set      _exclude
    cdef DiGraph  _graph  

    cpdef add(self,int node)
    cpdef remove(self, int node)

cdef class DiGraph:

    cdef dict _nodes
    cdef dict _edges_attrs
    cdef dict _nodes_attrs
    cdef str _name
    
    cpdef dict addNode(self,int node)
    cpdef dict addEdge(self,int node1, int node2)
    cpdef bint hasEdge(self, int node1, int node2)  
    cpdef dict getEdgeAttr(self, int node1, int node2,int edge=?)
    cpdef dict getNodeAttr(self, int node)
    cpdef int nodeCount(self)
    cpdef int edgeCount(self)
    cpdef str dot(self,nodePredicate=?, edgePredicate=?)
    cpdef str gml(self,nodePredicate=?, edgePredicate=?)
    cpdef deleteEdge(self, int node1, int node2,int edge=?)
    cpdef deleteNode(self, int node)
    cpdef DiGraph subgraph(self,str name, nodes)
    cpdef set connectedComponent(self,int node, nodePredicate=?, edgePredicate=?)
    


'''
Created on 28 sept. 2014

@author: coissac
'''
from orgasm import getOutput,getIndex, getSeeds
from orgasm.tango import restoreGraph, estimateFragmentLength, genesincontig,\
    scaffold, selectGoodComponent

__title__="Print some statistics about the assembling graph"

default_config = { 
                 }

def addOptions(parser):
    parser.add_argument(dest='orgasm:indexfilename',  metavar='index', 
                        help='index root filename (produced by the oa index command)')
    
    parser.add_argument(dest='orgasm:outputfilename',     metavar='<output>', 
                                                          nargs='?', 
                                                          default=None,
                        help='output prefix' )
    
    parser.add_argument('--back',             dest='orgasm:back', 
                                              metavar='<insert size>',
                                              type=int, 
                                              action='store', 
                                              default=None, 
                        help='the number of bases taken at the end of '
                             'contigs to jump with pared-ends [default: <estimated>]')

    


def run(config):

    logger=config['orgasm']['logger']
    output = getOutput(config)

    r = getIndex(config)
    ecoverage,x,newprobes = getSeeds(r,config)  
    
    asm = restoreGraph(output+'.oax',r,x)
    
    logger.info("Evaluate fragment length")
    
    meanlength,sdlength = estimateFragmentLength(asm)
    
    if config['orgasm']['back'] is not None:
        back = config['orgasm']['back']
    elif config['orgasm']['back'] is None and meanlength is not None:
        back = int(meanlength + 4 * sdlength)
        if back > 500:
            back=500
    else:
        back = 300

    cg = asm.compactAssembling(verbose=False)
    
    genesincontig(cg,r,x)

    scaffold(asm,
             cg,
             minlink=5,
             back=int(back),
             addConnectedLink=False,
             logger=logger)

    ccs = list(cg.connectedComponentIterator())
    gcc = selectGoodComponent(cg)
    gnode=set()
    for cc in gcc:
        for e in cc:
            gnode.add(e[0])
            gnode.add(e[1])
    
    ucc = set()
    for cc in ccs:
        ccc = frozenset([-x for x in cc])
        if ccc not in ucc:
            ucc.add(frozenset(cc))
        

    output = open(output+".stats","w")

    print ("AssembledBasePairs:",len(asm)/2,file=output)
    print ("TotalConnectedComponents:",len(ccs),file=output)
    print ("UniqueConnectedComponents:",len(ucc),file=output)
    print ("GoodConnectedComponents:",len(ucc),file=output)
    print ("CompactNodes:",len(cg),file=output)
    print ("GoodCompactNodes:",len(gnode),file=output)
    print ("CompactEdges:",cg.edgeCount(),file=output)
    print ("GoodCompactEdges:",sum(len(x) for x in gcc),file=output)
    print ("FragmentMeanLength:",meanlength,file=output)
    print ("FragmentSdLength:",sdlength,file=output)
    
    
    

              



#cython: language_level=3, boundscheck=False, wraparound=False
'''
Created on 14 sept. 2009

@author: coissac
'''
######
#
# Import standard memory management function to improve
# efficiency of the alignment code
#
#

from ._dynamic cimport * 
    
cdef AlignMatrix* allocateMatrix(int hsize, int vsize,AlignMatrix *matrix=NULL):
    
    cdef int msize
    
    vsize+=1
    hsize+=1
    
    if matrix is NULL:
        matrix = <AlignMatrix*>malloc(sizeof(AlignMatrix))
        matrix.msize=0
        matrix.matrix=NULL
        
    msize = hsize * vsize 
    if msize > matrix.msize:
        matrix.msize = msize
        matrix.matrix = <AlignCell*>realloc(matrix.matrix, msize * sizeof(AlignCell))
        
    return matrix

cdef void freeMatrix(AlignMatrix* matrix):
    if matrix is not NULL:
        if matrix.matrix is not NULL:
            free(matrix.matrix)
        free(matrix)
        
cdef void resetMatrix(AlignMatrix* matrix):
    if matrix is not NULL:
        if matrix.matrix is not NULL:
            bzero(<void*>matrix.matrix, matrix.msize * sizeof(AlignCell))
     
    
cdef alignPath* allocatePath(long l1,long l2,alignPath* path=NULL):
    cdef long length=l1+l2
    
    if path is NULL:
        path = <alignPath*>malloc(sizeof(alignPath))
        path.length=0
        path.buffsize=0
        path.path=NULL
        
    if length > path.buffsize:
        path.buffsize=length
        path.path=<long*>realloc(path.path,sizeof(long)*length)
        
    path.length=0
    
    return path

cdef void reversePath(alignPath* path):
        cdef long i
        cdef long j
        
        j=path.length
        for i in range(path.length//2):
            j-=1
            path.path[i],path.path[j]=path.path[j],path.path[i]

cdef void freePath(alignPath* path):
    if path is not NULL:
        if path.path is not NULL:
            free(<void*>path.path)
        free(<void*>path)
  
    
cdef class EndGapFree:

    def __init__(self):
        self.sequenceChanged=True

        self.matrix=NULL
        self.hSeq=NULL
        self.vSeq=NULL
        self.path=NULL
        
        self.horizontalSeq=None
        self.verticalSeq=None
        
    cdef int allocate(self) except -1:
        
        assert self.horizontalSeq is not None,'Sequence A must be set'
        assert self.verticalSeq is not None,'Sequence B must be set'
        
        cdef long lenH=len(self.horizontalSeq)
        cdef long lenV=len(self.verticalSeq)

        self.matrix=allocateMatrix(lenH,lenV,self.matrix)
        return 0


    cdef bint matchScore(self,int h, int v):
        return self.hSeq[h-1]==self.vSeq[v-1]
        
    cdef inline int index(self, int x, int y):
        return (self.lhSeq+1) * y + x

    cdef long doAlignment(self) except? 0:
        cdef int i  # vertical index
        cdef int j  # horizontal index
        cdef int idx
        cdef int jump
        cdef int delta
        cdef long score
        cdef long scoremax
        cdef long jmax
        cdef int    path

        
        if self.sequenceChanged:
            self.allocate()
            self.reset()
            
            for j in range(1,self.lhSeq+1):
                idx = self.index(j,0)
                self.matrix.matrix[idx].score = -2 * j  # <-- Gap score
                self.matrix.matrix[idx].path  = j
                                
            for i in range(1,self.lvSeq+1):
                idx = self.index(0,i)
                self.matrix.matrix[idx].score = 0
                self.matrix.matrix[idx].path  = -i
                
            for i in range(1,self.lvSeq+1):
                for j in range(1,self.lhSeq+1):
                    
                    # 1 - came from diagonal
                    idx = self.index(j-1,i-1)
                    # print "computing cell : %d,%d --> %d/%d" % (j,i,self.index(j,i),self.matrix.msize),
                    scoremax = self.matrix.matrix[idx].score + \
                               self.matchScore(j,i)
                    path = 0

                    # print "so=%f sd=%f sm=%f" % (self.matrix.matrix[idx].score,self.matchScore(j,i),scoremax),

                    # 2 - open horizontal gap
                    idx = self.index(j-1,i)
                    score = self.matrix.matrix[idx].score - 2 # <-- Gap score
                    if score > scoremax : 
                        scoremax = score
                        path = self.matrix.matrix[idx].path
                        if path >=0:
                            path+=1
                        else: 
                            path=+1
                    
                    # 3 - open vertical gap
                    idx = self.index(j,i-1)
                    score = self.matrix.matrix[idx].score - 2 # <-- Gap score
                    if score > scoremax : 
                        scoremax = score
                        path = self.matrix.matrix[idx].path
                        if path <=0:
                            path-=1
                        else:
                            path=-1
                                                    
                    idx = self.index(j,i)
                    self.matrix.matrix[idx].score = scoremax
                    self.matrix.matrix[idx].path  = path 
                                        
            scoremax=0
            jmax=self.lhSeq
            
            for j in range(self.lhSeq,1,-1):
                idx = self.index(j,self.lvSeq)
                if self.matrix.matrix[idx].score > scoremax:
                    scoremax=self.matrix.matrix[idx].score
                    jmax=j
#                    print j,scoremax,jmax
                    
            idx = self.index(self.lhSeq,self.lvSeq)
            self.matrix.matrix[idx].score=scoremax
            self.matrix.matrix[idx].path=self.lhSeq - jmax

        self.sequenceChanged=False

        idx = self.index(self.lhSeq,self.lvSeq)
        return self.matrix.matrix[idx].score
                   
    cdef void backtrack(self):
#        cdef list path=[]
        cdef int i
        cdef int j 
        cdef int p
        
        self.doAlignment()
        i=self.lvSeq
        j=self.lhSeq
        self.path=allocatePath(i,j,self.path)
        
        while (i or j):
            p=self.matrix.matrix[self.index(j,i)].path
            self.path.path[self.path.length]=p
            self.path.length+=1
#            path.append(p)
            if p==0:
                i-=1
                j-=1
            elif p < 0:
                i+=p
            else:
                j-=p
                
#        print path
        
    property seqA:
            def __get__(self):
                return self.horizontalSeq
            
            def __set__(self, seq):
                self.sequenceChanged=True
                self.horizontalSeq=seq
                self.hSeq=<char*>seq
                self.lhSeq=len(seq)
                
    property seqB:
            def __get__(self):  # @DuplicatedSignature
                return self.verticalSeq
            
            def __set__(self, seq):  # @DuplicatedSignature
                self.sequenceChanged=True
                self.verticalSeq=seq
                self.vSeq=<char*>seq
                self.lvSeq=len(seq)
                
    property score:
        def __get__(self):
            return self.doAlignment()
                    
    cdef void reset(self):
        self.sequenceChanged=True
        resetMatrix(self.matrix)
        
    
    cdef void clean(self):
        freeMatrix(self.matrix)
        freePath(self.path)
        
    def __dealloc__(self):
        self.clean()
        
    def __call__(self):
        cdef list hgaps=[]
        cdef list vgaps=[]
        cdef list b
        cdef int  hp=0
        cdef int  vp=0
        cdef int  lenh=0
        cdef int  lenv=0
        cdef int  h,v,p
        cdef int  i
        cdef object ali
        cdef long score
        
        if self.sequenceChanged:
            score = self.doAlignment()

        self.backtrack()
        for i in range(self.path.length-1,-1,-1):
            p=self.path.path[i]
            if p==0:
                hp+=1
                vp+=1
                lenh+=1
                lenv+=1
            elif p>0:
                hp+=p
                lenh+=p
                vgaps.append([vp,p])
                vp=0
            else:
                vp-=p
                lenv-=p
                hgaps.append([hp,-p])
                hp=0
            
        if hp:
            hgaps.append([hp,0])
        if vp:
            vgaps.append([vp,0])
                    
        return hgaps,vgaps
  
        
            
                